<?php
namespace webaware\eway_payment_gateway;

if (!defined('ABSPATH')) {
	exit;
}

/**
* Classes for dealing with eWAY payments
*
* NB: for testing, the only card number seen as valid is '4444333322221111'
*
* @link https://www.eway.com.au/developers/api/direct-payments
* @link https://www.eway.com.au/developers/api/beagle-lite
*/

/**
* Class for dealing with an eWAY payment
*/
class EwayLegacyAPI {

	#region members

	// environment / website specific members
	/**
	* default FALSE, use eWAY sandbox unless set to TRUE
	* @var boolean
	*/
	public $isLiveSite;

	/**
	* default TRUE, whether to validate the remote SSL certificate
	* @var boolean
	*/
	public $sslVerifyPeer;

	// payment specific members
	/**
	* account name / email address at eWAY
	* @var string max. 8 characters
	*/
	public $accountID;

	/**
	* an invoice reference to track by (NB: see transactionNumber which is intended for invoice number or similar)
	* @var string max. 50 characters
	*/
	public $invoiceReference;

	/**
	* description of what is being purchased / paid for
	* @var string max. 10000 characters
	*/
	public $invoiceDescription;

	/**
	* total amount of payment, in dollars and cents as a floating-point number (will be converted to just cents for transmission)
	* @var float
	*/
	public $amount;

	/**
	* customer's first name
	* @var string max. 50 characters
	*/
	public $firstName;

	/**
	* customer's last name
	* @var string max. 50 characters
	*/
	public $lastName;

	/**
	* customer's email address
	* @var string max. 50 characters
	*/
	public $emailAddress;

	/**
	* customer's address line 1
	* @var string max. 50 characters
	*/
	public $address1;

	/**
	* customer's address line 2
	* @var string max. 50 characters
	*/
	public $address2;

	/**
	* customer's postcode
	* @var string max. 6 characters
	*/
	public $postcode;

	/**
	* customer's suburb/city/town
	* @var string max. 50 characters
	*/
	public $suburb;

	/**
	* customer's state/province
	* @var string max. 50 characters
	*/
	public $state;

	/**
	* country name
	* @var string
	*/
	public $countryName;

	/**
	* country code for billing address
	* @var string 2 characters
	*/
	public $country;

	/**
	* name on credit card
	* @var string max. 50 characters
	*/
	public $cardHoldersName;

	/**
	* credit card number, with no spaces
	* @var string max. 20 characters
	*/
	public $cardNumber;

	/**
	* month of expiry, numbered from 1=January
	* @var integer max. 2 digits
	*/
	public $cardExpiryMonth;

	/**
	* year of expiry
	* @var integer will be truncated to 2 digits, can accept 4 digits
	*/
	public $cardExpiryYear;

	/**
	* CVN (Creditcard Verification Number) for verifying physical card is held by buyer
	* @var string max. 3 or 4 characters (depends on type of card)
	*/
	public $cardVerificationNumber;

	/**
	* eWAYTrxnNumber - This value is returned to your website.
	*
	* You can pass a unique transaction number from your site. You can update and track the status of a transaction when eWAY
	* returns to your site.
	*
	* NB. This number is returned as 'ewayTrxnReference', member transactionReference of EwayResponse.
	*
	* @var string max. 16 characters
	*/
	public $transactionNumber;

	/**
	* optional additional information for use in shopping carts, etc.
	* @var array[string] max. 255 characters, up to 3 elements
	*/
	public $options = [];

	/**
	* Beagle: IP address of purchaser (from REMOTE_ADDR)
	* @var string max. 15 characters
	*/
	public $customerIP;

	#endregion

	#region constants

	/** host for the eWAY Real Time API with CVN verification in the developer sandbox environment */
	const REALTIME_CVN_API_SANDBOX = 'https://www.eway.com.au/gateway_cvn/xmltest/testpage.asp';
	/** host for the eWAY Real Time API with CVN verification in the production environment */
	const REALTIME_CVN_API_LIVE = 'https://www.eway.com.au/gateway_cvn/xmlpayment.asp';
	/** host for the eWAY Beagle API in the developer sandbox environment */
	const REALTIME_BEAGLE_API_SANDBOX = 'https://www.eway.com.au/gateway_cvn/xmltest/BeagleTest.aspx';
	/** host for the eWAY Beagle API in the production environment */
	const REALTIME_BEAGLE_API_LIVE = 'https://www.eway.com.au/gateway_cvn/xmlbeagle.asp';

	#endregion

	/**
	* populate members with defaults, and set account and environment information
	*
	* @param string $accountID eWAY account ID
	* @param boolean $isLiveSite running on the live (production) website
	*/
	public function __construct($accountID, $isLiveSite = false) {
		$this->sslVerifyPeer	= true;
		$this->isLiveSite		= $isLiveSite;
		$this->accountID		= $accountID;
	}

	/**
	* process a payment against eWAY; throws exception on error with error described in exception message.
	*/
	public function processPayment() {
		$this->validate();
		$xml = $this->getPaymentXML();
		return $this->sendPayment($xml);
	}

	/**
	* validate the data members to ensure that sufficient and valid information has been given
	*/
	private function validate() {
		$errors = [];

		if (strlen($this->accountID) === 0) {
			$errors[] = __('CustomerID cannot be empty', 'eway-payment-gateway');
		}
		if (!is_numeric($this->amount) || $this->amount <= 0) {
			$errors[] = __('amount must be given as a number in dollars and cents', 'eway-payment-gateway');
		}
		else if (!is_float($this->amount)) {
			$this->amount = (float) $this->amount;
		}
		if (strlen($this->cardHoldersName) === 0) {
			$errors[] = __('card holder name cannot be empty', 'eway-payment-gateway');
		}
		if (strlen($this->cardNumber) === 0) {
			$errors[] = __('card number cannot be empty', 'eway-payment-gateway');
		}

		// make sure that card expiry month is a number from 1 to 12
		if (!is_int($this->cardExpiryMonth)) {
			if (strlen($this->cardExpiryMonth) === 0) {
				$errors[] = __('card expiry month cannot be empty', 'eway-payment-gateway');
			}
			elseif (!ctype_digit($this->cardExpiryMonth)) {
				$errors[] = __('card expiry month must be a number between 1 and 12', 'eway-payment-gateway');
			}
			else {
				$this->cardExpiryMonth = intval($this->cardExpiryMonth);
			}
		}
		if (is_int($this->cardExpiryMonth)) {
			if ($this->cardExpiryMonth < 1 || $this->cardExpiryMonth > 12) {
				$errors[] = __('card expiry month must be a number between 1 and 12', 'eway-payment-gateway');
			}
		}

		// make sure that card expiry year is a 2-digit or 4-digit year >= this year
		if (!is_int($this->cardExpiryYear)) {
			if (strlen($this->cardExpiryYear) === 0) {
				$errors[] = __('card expiry year cannot be empty', 'eway-payment-gateway');
			}
			elseif (!ctype_digit($this->cardExpiryYear)) {
				$errors[] = __('card expiry year must be a two or four digit year', 'eway-payment-gateway');
			}
			else {
				$this->cardExpiryYear = intval($this->cardExpiryYear);
			}
		}
		if (is_int($this->cardExpiryYear)) {
			$thisYear = intval(date_create()->format('Y'));
			if ($this->cardExpiryYear < 0 || $this->cardExpiryYear >= 100 && $this->cardExpiryYear < 2000 || $this->cardExpiryYear > $thisYear + 20) {
				$errors[] = __('card expiry year must be a two or four digit year', 'eway-payment-gateway');
			}
			else {
				if ($this->cardExpiryYear > 100 && $this->cardExpiryYear < $thisYear) {
					$errors[] = __("card expiry can't be in the past", 'eway-payment-gateway');
				}
				else if ($this->cardExpiryYear < 100 && $this->cardExpiryYear < ($thisYear - 2000)) {
					$errors[] = __("card expiry can't be in the past", 'eway-payment-gateway');
				}
			}
		}

		if (count($errors) > 0) {
			throw new EwayPaymentsException(implode("\n", $errors));
		}
	}

	/**
	* create XML request document for payment parameters
	*
	* @return string
	*/
	public function getPaymentXML() {
		// aggregate street, city, state, country into a single string
		$parts = [$this->address1, $this->address2, $this->suburb, $this->state, $this->countryName];
		$address = implode(', ', array_filter($parts, 'strlen'));

		$xml = new \XMLWriter();
		$xml->openMemory();
		$xml->startDocument('1.0', 'UTF-8');
		$xml->startElement('ewaygateway');

		$xml->writeElement('ewayCustomerID', $this->accountID);
		$xml->writeElement('ewayTotalAmount', number_format($this->amount * 100, 0, '', ''));
		$xml->writeElement('ewayCustomerFirstName', $this->firstName);
		$xml->writeElement('ewayCustomerLastName', $this->lastName);
		$xml->writeElement('ewayCustomerEmail', $this->emailAddress);
		$xml->writeElement('ewayCustomerAddress', $address);
		$xml->writeElement('ewayCustomerPostcode', $this->postcode);
		$xml->writeElement('ewayCustomerInvoiceDescription', $this->invoiceDescription);
		$xml->writeElement('ewayCustomerInvoiceRef', $this->invoiceReference);
		$xml->writeElement('ewayCardHoldersName', $this->cardHoldersName);
		$xml->writeElement('ewayCardNumber', $this->cardNumber);
		$xml->writeElement('ewayCardExpiryMonth', sprintf('%02d', $this->cardExpiryMonth));
		$xml->writeElement('ewayCardExpiryYear', sprintf('%02d', $this->cardExpiryYear % 100));
		$xml->writeElement('ewayTrxnNumber', $this->transactionNumber);
		$xml->writeElement('ewayOption1', empty($this->option[0]) ? '' : $this->option[0]);
		$xml->writeElement('ewayOption2', empty($this->option[1]) ? '' : $this->option[1]);
		$xml->writeElement('ewayOption3', empty($this->option[2]) ? '' : $this->option[2]);
		$xml->writeElement('ewayCVN', $this->cardVerificationNumber);

		// Beagle data
		if (!empty($this->country) && $this->accountID !== EWAY_PAYMENTS_TEST_CUSTOMER) {
			if (empty($this->customerIP)) {
				$this->customerIP = get_customer_IP($this->isLiveSite);
			}
			$xml->writeElement('ewayCustomerIPAddress', $this->customerIP);
			$xml->writeElement('ewayCustomerBillingCountry', $this->country);
		}

		$xml->endElement();		// ewaygateway

		return $xml->outputMemory();
	}

	/**
	* send the eWAY payment request and retrieve and parse the response
	*
	* @return EwayResponse
	* @param string $xml eWAY payment request as an XML document, per eWAY specifications
	*/
	private function sendPayment($xml) {
		// select endpoint URL, use sandbox if not from live website
		if (!empty($this->country) && $this->accountID !== EWAY_PAYMENTS_TEST_CUSTOMER) {
			// use Beagle anti-fraud endpoints
			$url = $this->isLiveSite ? self::REALTIME_BEAGLE_API_LIVE : self::REALTIME_BEAGLE_API_SANDBOX;
		}
		else {
			// normal Direct Payments endpoints with CVN verification
			$url = $this->isLiveSite ? self::REALTIME_CVN_API_LIVE : self::REALTIME_CVN_API_SANDBOX;
		}

		// execute the cURL request, and retrieve the response
		try {
			$responseXML = self::xmlPostRequest($url, $xml, $this->sslVerifyPeer);
		}
		catch (EwayPaymentsException $e) {
			throw new EwayPaymentsException("Error posting eWAY payment to $url: " . $e->getMessage());
		}

		$response = new EwayResponseLegacyDirect();
		$response->loadResponseXML($responseXML);
		return $response;
	}

	/**
	* send XML data via HTTP and return response
	* @param string $url
	* @param string $data
	* @param bool $sslVerifyPeer whether to validate the SSL certificate
	* @return string $response
	* @throws EwayPaymentsException
	*/
	protected static function xmlPostRequest($url, $data, $sslVerifyPeer = true) {
		// send data via HTTPS and receive response
		$response = wp_remote_post($url, [
			'user-agent'	=> 'WordPress/eWAY Payment Gateway ' . EWAY_PAYMENTS_VERSION,
			'sslverify'		=> $sslVerifyPeer,
			'timeout'		=> 60,
			'headers'		=> ['Content-Type' => 'text/xml; charset=utf-8'],
			'body'			=> $data,
		]);

		if (is_wp_error($response)) {
			throw new EwayPaymentsException($response->get_error_message());
		}

		// error code returned by request
		$code = wp_remote_retrieve_response_code($response);
		if ($code !== 200) {
			$msg = wp_remote_retrieve_response_message($response);

			if (empty($msg)) {
				/* translators: %s = the error code */
				$msg = sprintf(__('Error posting eWAY request: %s', 'eway-payment-gateway'), $code);
			}
			else {
				/* translators: 1. the error code; 2. the error message */
				$msg = sprintf(__('Error posting eWAY request: %1$s, %2$s', 'eway-payment-gateway'), $code, $msg);
			}
			throw new EwayPaymentsException($msg);
		}

		return wp_remote_retrieve_body($response);
	}

}

/**
* Class for dealing with an eWAY payment response
*/
class EwayResponseLegacyDirect extends EwayResponse {

	#region members

	/**
	* bank authorisation code
	* @var string
	*/
	public $AuthorisationCode;

	/**
	* array of codes describing the result (including Beagle failure codes)
	* @var array
	*/
	public $ResponseMessage;

	/**
	* eWAY transacation ID
	* @var string
	*/
	public $TransactionID;

	/**
	* eWAY transaction status: true for success
	* @var boolean
	*/
	public $TransactionStatus;

	/**
	* Beagle fraud detection score
	* @var string
	*/
	public $BeagleScore;

	/**
	* payment details object
	* @var object
	*/
	public $Payment;

	/**
	* a list of errors -- just the one for the Direct API
	* @var
	*/
	public $Errors;

	#endregion

	/**
	* load eWAY response data as XML string
	*
	* @param string $response eWAY response as a string (hopefully of XML data)
	*/
	public function loadResponseXML($response) {
		// make sure we actually got something from eWAY
		if (strlen($response) === 0) {
			throw new EwayPaymentsException(__('eWAY payment request returned nothing; please check your card details', 'eway-payment-gateway'));
		}

		// prevent XML injection attacks, and handle errors without warnings
		$oldDisableEntityLoader = libxml_disable_entity_loader(true);
		$oldUseInternalErrors = libxml_use_internal_errors(true);

		try {
			$xml = simplexml_load_string($response);
			if ($xml === false) {
				$errmsg = '';
				foreach (libxml_get_errors() as $error) {
					$errmsg .= $error->message;
				}
				throw new Exception($errmsg);
			}

			$this->AuthorisationCode			= (string) $xml->ewayAuthCode;
			$this->ResponseMessage				= [];
			$this->TransactionStatus			= (strcasecmp((string) $xml->ewayTrxnStatus, 'true') === 0);
			$this->TransactionID				= (string) $xml->ewayTrxnNumber;
			$this->BeagleScore					= (string) $xml->ewayBeagleScore;
			$this->Errors						= ['ERROR' => (string) $xml->ewayTrxnError];

			// if we got an amount, convert it back into dollars.cents from just cents
			$this->Payment						= new \stdClass;
			$this->Payment->TotalAmount			= empty($xml->ewayReturnAmount) ? null : floatval($xml->ewayReturnAmount) / 100.0;
			$this->Payment->InvoiceReference	= (string) $xml->ewayTrxnReference;

			// restore old libxml settings
			libxml_disable_entity_loader($oldDisableEntityLoader);
			libxml_use_internal_errors($oldUseInternalErrors);
		}
		catch (Exception $e) {
			// restore old libxml settings
			libxml_disable_entity_loader($oldDisableEntityLoader);
			libxml_use_internal_errors($oldUseInternalErrors);

			throw new EwayPaymentsException(sprintf(__('Error parsing eWAY response: %s', 'eway-payment-gateway'), $e->getMessage()));
		}
	}

	/**
	* get 'invalid response' message for this response class
	* @return string
	*/
	protected function getMessageInvalid() {
		return __('Invalid response from eWAY for legacy XML Direct payment', 'eway-payment-gateway');
	}

}
